import click
import os
from . import __version__
from .hexint import HexInt
from .utils import check_pid, get_offest, virtual_to_physical
from .constant import *
from .row import Row


# Convert virtual memory address to physical memory address
@click.command()
@click.option(
    "--pid", "-p", type=int, default=os.getpid(), help="the process id to convert"
)
@click.option(
    "--virtual_address",
    "-v",
    type=HexInt(),
    required=True,
    help="the virtual address to convert",
)
def convert(pid: int, virtual_address: int):
    print(f"pid = {pid}, va = {hex(virtual_address)}")
    assert check_pid(pid), f"process {pid} is not exists."

    offset = get_offest(virtual_address)
    with open(f"/proc/{pid}/pagemap", "rb") as f:
        f.seek(offset)
        row = Row(f.read(ROW_SIZE))
        print(row)
        if not row.present:
            print(f"{hex(virtual_address)} is not present.")
            return
        physical_address = virtual_to_physical(virtual_address, row.pfn)
        print(f"{hex(virtual_address)} -> {hex(physical_address)}")
        f.close()


# Count the actual number of virtual memory pages currently occupied by the process,
# the number of physical memory pages, and the number of virtual memory pages that
# have been swapped to external memory.
@click.command()
@click.option(
    "--pid", "-p", type=int, default=os.getpid(), help="the process id to convert"
)
def show(pid: int):
    virtual_page_cnt = 0
    physical_page_cnt = 0
    swapped_cnt = 0
    print(f"pid = {pid}")
    with open(f"/proc/{pid}/maps", "r") as maps:
        with open(f"/proc/{pid}/pagemap", "rb") as pagemap:
            for line in maps.readlines():
                begin, end = map(lambda x: int(x, 16),
                                 (line.split()[0]).split("-"))
                for offset in map(get_offest, range(begin, end, PAGE_SIZE)):
                    pagemap.seek(offset, 0)
                    virtual_page_cnt += 1
                    row = Row(pagemap.read(ROW_SIZE))
                    if row.present:
                        physical_page_cnt += 1
                    if row.swapped:
                        swapped_cnt += 1
            pagemap.close()
        maps.close()
    print(
        f"virtual pages: {virtual_page_cnt}, physical pages: {physical_page_cnt}, swapped pages: {swapped_cnt}"
    )


@click.version_option(__version__)
@click.group()
def run():
    pass


run.add_command(convert)
run.add_command(show)


def main():
    run()



import sys


class Row(object):
    def __init__(self, data: bytes):
        data = int.from_bytes(data, sys.byteorder)
        self.pfn = data & ((1 << 55) - 1)
        self.file_or_shared = (data >> 61) & 1 == 1
        self.soft_dirty = (data >> 55) & 1 == 1
        self.swapped = (data >> 62) & 1 == 1
        self.present = (data >> 63) & 1 == 1

    def __str__(self) -> str:
        return f'swapped: {self.present}, present: {self.swapped}, file/shared: {self.file_or_shared}, soft-dirty: {self.soft_dirty}, pfn: {hex(self.pfn)}'



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pagemap']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1,<9.0']

entry_points = \
{'console_scripts': ['pagemap = pagemap.entry:main']}

setup_kwargs = {
    'name': 'pagemap',
    'version': '0.1.0',
    'description': 'A tool for virtual memory.',
    'long_description': '# pagemap\n\nA tool for linux memory stats.\n',
    'author': 'fffzlfk',
    'author_email': '1319933925qq@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import click


class HexInt(click.ParamType):
    name = "hex integer"

    def convert(self, value, param, ctx):
        try:
            if value[:2].lower() == "0x":
                return int(value[2:], 16)
        except ValueError:
            self.fail(f"{value!r} is not a valid hex integer", param, ctx)



import os
from .constant import *


def check_pid(pid):
    """ check for the existence of a unix pid. """
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    else:
        return True


def get_offest(virtual_address: int) -> int:
    return virtual_address // PAGE_SIZE * ROW_SIZE


def virtual_to_physical(virtual_address: int, pfn: int) -> int:
    page_offset = virtual_address % PAGE_SIZE
    return pfn * PAGE_SIZE + page_offset
